/*
 * Copyright (c) 2014 Cloudera, Inc. All rights reserved.
 */

package com.cloudera.keytrustee;

import com.cloudera.keytrustee.impl.ClientFactoryImpl;
import com.cloudera.keytrustee.impl.FileTokenStore;
import com.cloudera.keytrustee.util.ConfigUpdate;
import com.cloudera.keytrustee.util.Environment;
import org.apache.commons.io.IOUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.crypto.key.KeyProvider;
import org.apache.hadoop.crypto.key.KeyProviderFactory;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.net.NetUtils;
import org.apache.hadoop.security.Groups;
import org.apache.hadoop.security.ProviderUtils;
import org.apache.hadoop.security.UserGroupInformation;
import org.apache.hadoop.security.authorize.AuthorizationException;

import com.google.common.base.Preconditions;

import javax.crypto.Cipher;
import javax.xml.ws.http.HTTPException;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.security.NoSuchAlgorithmException;
import java.text.MessageFormat;
import java.text.ParseException;
import java.util.*;
import java.util.concurrent.TimeoutException;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

/**
 * KeyProvider that uses KeyTrustee opaque object store.
 */
public class TrusteeKeyProvider extends KeyProvider {

  public static final String SCHEME_NAME = "keytrustee";
  public static final String DEPOSIT_STATE_ENABLED = "Enabled";
  public static final String BIT_LENGTH_ATTRIBUTE = "_" + SCHEME_NAME + "_bit_length";
  public static final String CIPHER_ATTRIBUTE = "_" + SCHEME_NAME + "cipher";
  public static final String DESCRIPTION_ATTRIBUTE = "_" + SCHEME_NAME + "description";
  public static final String VERSIONS_ATTRIBUTE = "_" + SCHEME_NAME + "versions";
  public static final String CREATED_ATTRIBUTE = "_" + SCHEME_NAME + "created";

  private static final Log LOG = LogFactory.getLog(TrusteeKeyProvider.class);
  private static final DepositSearch QUERY_ALL_DEPOSITS = new DepositSearch.Builder().build();
  private static final DepositSearch QUERY_ALL_ENABLED_DEPOSITS = new DepositSearch.Builder().enabled(true).build();

  private final File keyTrusteeConfDir;

  private final ClientConnection keyTrusteeConnectionProxy;
  private final boolean useKeyTrusteeConnectionPool;

  private final boolean strongLocking;
  private final Lock readLock;
  private final Lock writeLock;

  private FileTokenStore tokenStore = null;
  //private URL keyTrusteeURL;
  Environment env;
  private URI[] uris;
  private Groups groupsMapping;

  /**
   * Creates an instance of the KeyProvider
   * KeyProvider expects to find keyTrustee configuration directory on the local file system
   * at the location specified in the uri.
   *
   * @param uri  path specified using the keytrustee scheme (e.g. keytrustee://file@/${user.home}/.ztrustee/)
   * @param conf Hadoop configuration. Can contain properties/env. vars (e.g. KEYTRUSTEE_SERVER_HOST,
   *             KEYTRUSTEE_SERVER_ORG, KEYTRUSTEE_SERVER_AUTH, KEYTRUSTEE_SERVER_CERT_FILE,
   *             KEYTRUSTEE_KEY_LENGTH and KEYTRUSTEE_CONF_DIR)
   */
  public TrusteeKeyProvider(URI uri, Configuration conf) throws IOException {
    this(null, uri, conf);
  }

  public TrusteeKeyProvider(Environment env, URI uri, Configuration conf) throws IOException {
    super(conf);
    // This is a temporary variable rather than a member variable so that Configuration
    // object can be garbage collected freely rather than living for the life of this object.
    TrusteeKeyProviderConfiguration tkpConfig = new TrusteeKeyProviderConfiguration(env, uri, conf);

    boolean isNewClientSetup = false;
    
    uris = tkpConfig.getSpecifiedProviderURIs();
    keyTrusteeConfDir = tkpConfig.getKeyTrusteeConfDir();

    groupsMapping = Groups.getUserToGroupsMappingService(conf);

    int maxKeyLength = -1;
    try {
      maxKeyLength = Cipher.getMaxAllowedKeyLength("AES");
    } catch (NoSuchAlgorithmException e) {
      throw new Error("No support for AES algorithm. Unexpected.");
    }

    if (!(maxKeyLength > 128)) {
      throw new Error("Max key length is: " + maxKeyLength + ". JCE Unlimited Strength Policy files are not detected. " +
      "JCE Unlimited Strength policy files are required to use Trustee key provider.");
    }

    createInitialClient(tkpConfig, isNewClientSetup);

    try {
      updateClientConfiguration(tkpConfig, new ClientFactoryImpl());
    } catch(KeyTrusteeException e) {
      if (LOG.isErrorEnabled()) {
        LOG.error("Attempt to verify and/or update client security settings failed. Continuing with existing settings.");
      }
    }

    this.useKeyTrusteeConnectionPool = tkpConfig.isUseKeyTrusteeConnectionPool();
    this.strongLocking = tkpConfig.isStrongLocking();

    try {
      if (!this.useKeyTrusteeConnectionPool) {
        ClientConnection keyTrusteeConnection = getConnection(new ClientFactoryImpl());
        //this.keyTrusteeConnectionPool = null;
        this.keyTrusteeConnectionProxy = ClientConnectionInvocationHandler.createClientConnection(keyTrusteeConnection);
      } else {
        //this.keyTrusteeConnection = null;
        TrusteeClientConnectionPool keyTrusteeConnectionPool = initializeConnectionPool(keyTrusteeConfDir, tkpConfig);
        this.keyTrusteeConnectionProxy = ClientConnectionInvocationHandler.createClientConnection(keyTrusteeConnectionPool);
      }
    } catch (KeyTrusteeException e) {
      sendDebugInformationToLog("TrusteeKeyProvider(Environment, URI, Configuration)", e);
      throw new IOException("Could not open KeyTrustee client.", e);
    }

    ReadWriteLock lock = new ReentrantReadWriteLock(true);
    readLock = lock.readLock();
    writeLock = lock.writeLock();

    if (LOG.isDebugEnabled()) {
      LOG.debug("keytrustee conf dir: " + keyTrusteeConfDir);
    }
  }

   /*
    * Helper method for constructor only.
    */
  private void createInitialClient(TrusteeKeyProviderConfiguration tkpConfig, boolean isNewClientSetup) throws IOException {
    ClientFactory keyTrusteeClientFactory = new ClientFactoryImpl();
    try {
      ClientInfo keyTrusteeClient;

      if (!(keyTrusteeClientFactory.clientExists(keyTrusteeConfDir) &&
              isRegistered(keyTrusteeClientFactory, keyTrusteeConfDir))) {
        int keyLength = tkpConfig.getKeyLength();

        // TODO: When KT-1478 is implemented, set timeout in config like other properties rather than setting sys prop
        System.setProperty(URLConnector.KEYTRUSTEE_API_READ_TIMEOUT_KEY,
                String.valueOf(tkpConfig.getHttpReadConnectionTimeout()));

        if (keyTrusteeClientFactory.clientExists(keyTrusteeConfDir)) {
          keyTrusteeClient = keyTrusteeClientFactory.openClient(keyTrusteeConfDir);
        } else {
          keyTrusteeClient = keyTrusteeClientFactory.createNewClient(KeyLength.of(keyLength), keyTrusteeConfDir);
          isNewClientSetup = true;
        }
        
        //TODO: Start Factor out this code for use in updateClientConfiguration (if needed)
        TrusteeKeyProviderConfiguration.Register register = tkpConfig.getRegisterCommand().invoke();

        register.invariant(); // check whether any required properties are missing

        try {
          FailoverServerInfo.Builder builder =
                  (FailoverServerInfo.Builder)new FailoverServerInfo.Builder()
                          .hostnames(Arrays.asList(register.getHostnames())).setRoundRobin(register.isRoundRobin())
                          .certificate(register.getCertPath()).sslInsecure(register.isInsecure())
                          .protocol(register.getProtocol()).tokenSync(register.isTokenSync());
            
          if (null != register.getHkpPort()) builder.hkpPort(register.getHkpPort());
          if (null != register.getKtsPort()) builder.ktsPort(register.getKtsPort());
          if (null != register.getHkpSsl()) builder.singleServerPort(register.getHkpSsl());

          FailoverServerInfo serverInfo = (FailoverServerInfo)builder.build();

          ClientConnection conn =
                  register.isTokenSync() ? keyTrusteeClientFactory.connect(keyTrusteeClient, serverInfo,
                          new FileTokenStore(keyTrusteeConfDir, serverInfo)) : keyTrusteeClientFactory.connect(
                          keyTrusteeClient, serverInfo);
          if (null != register.getOrg() && !register.getOrg().isEmpty()) {
            ActivationRequest activationRequest = new ActivationRequest.Builder()
                    .org(register.getOrg()).auth(register.getAuth()).cert(register.getCertPath()).build();
            conn.registerAndActivate(activationRequest);
            // keyTrusteeConnection = conn;
            keyTrusteeClient.saveConfig(keyTrusteeConfDir);
          }
          //TODO: End Factor out this code for use in updateClientConfiguration
          
        } catch (Exception e) {
          if (isNewClientSetup) removeConfDir(keyTrusteeConfDir);
          throw new IOException(e);
        }
      }
    } catch (KeyTrusteeException e) {
      sendDebugInformationToLog("createInitialClient", e);
      throw new IOException("TrusteeKeyProvider initialization failed.", e);
    }
  }
  /*
   * Helper method for constructor only.
   */
  private ClientConnection getConnection(ClientFactory keyTrusteeClientFactory) throws KeyTrusteeException, IOException {
    ClientInfo keyTrusteeClient = keyTrusteeClientFactory.openClient(keyTrusteeConfDir);
    ServerInfo keyTrusteeServer = keyTrusteeClient.getServerInfo();
    ClientConnection keyTrusteeConnection = null;
    if (keyTrusteeServer.isTokenSync()) {
      this.tokenStore = new FileTokenStore(keyTrusteeConfDir, keyTrusteeServer);
      keyTrusteeConnection = keyTrusteeClientFactory.connect(keyTrusteeClient, keyTrusteeServer, tokenStore);
    } else {
      keyTrusteeConnection = keyTrusteeClientFactory.connect(keyTrusteeClient, keyTrusteeServer);
    }
    return keyTrusteeConnection;
  }
  /*
   * Helper method for constructor only.
   */
  private void updateClientConfiguration(TrusteeKeyProviderConfiguration tkpConfig, ClientFactory keyTrusteeClientFactory)
          throws KeyTrusteeException, IOException {
    ClientInfo keyTrusteeClient = keyTrusteeClientFactory.openClient(keyTrusteeConfDir);
    boolean updatedClientConfig = tkpConfig.updateClientConfiguration(keyTrusteeClient);
    if (updatedClientConfig) {
      keyTrusteeClient.saveConfig(keyTrusteeConfDir);
      if (LOG.isWarnEnabled()) {
        LOG.warn("Client configuration has been updated in the keytrustee client configuration at " + keyTrusteeConfDir);
      }
    }
  }

  /**
   * Create a new key. The given key must not already exist.
   *
   * @param name     the base name of the key (KeyTrustee handle name)
   * @param material the key material for the first version of the key.
   * @param options  the options for the new key.
   * @return the version name of the first version of the key.
   * @throws IOException
   */
  @Override
  public KeyVersion createKey(String name, byte[] material, Options options) throws IOException {
    if (LOG.isDebugEnabled()) LOG.debug("Entering createKey method.");
    Preconditions.checkNotNull(name, "Null name passed to createKey.");
    Preconditions.checkNotNull(material, "Null material passed to createKey.");
    Preconditions.checkNotNull(options, "Null options passed to createKey.");
    try {
      writeLock.lock();
      KeyExistsResult keyExistsResult = keyExists(name);
      if (keyExistsResult.isExisting()) {
        String message = "Key with name \"" + name + "\" already exists in \"" + this + ". ";
        String detail = keyExistsResult.getReason();
        throw new DuplicateKeyException(message + detail);
      }
      Deposit deposit = createHandleDeposit(name, material, options);
      String versionName = null;
      try {
        versionName = putDeposit(deposit);
      } catch (KeyTrusteeException e) {
        sendDebugInformationToLog("createKey", e);
        throw new IOException("Creation of key named " + name + " failed.", e);
      }
      return new KeyTrusteeKeyVersion(name, versionName, material);
    } finally {
      writeLock.unlock();
      if (LOG.isDebugEnabled()) LOG.debug("Exiting createKey method.");
    }
  }

    /**
   * Get the current version of the key, which should be used for encrypting new
   * data.
   *
   * @param name the base name of the key (KeyTrustee handle name)
   * @return the version name (KeyTrustee UUID) of the current version of the key
   * or null if the key version doesn't exist
   * @throws IOException
   */
  @Override
  public KeyVersion getCurrentKey(String name) throws IOException {
    if (LOG.isDebugEnabled()) LOG.debug("Entering getCurrentKey method.");
    Preconditions.checkNotNull(name, "Null name passed to getCurrentKey.");

    KeyVersion keyVersion = null;

    try {
      List<KeyVersion> keyVersions = getKeyVersions(name);
      if (!keyVersions.isEmpty()) keyVersion = keyVersions.get(0);
    } catch (KeyNotFoundException e) {
      // ignore; return null keyVersion
    } finally {
      if (LOG.isDebugEnabled()) LOG.debug("Exiting getCurrentKey method.");
    }
    return keyVersion;
  }

  /**
   * Get the key material for a specific version of the key. This method is used
   * when decrypting data.
   *
   * @param versionName KeyTrustee UUID for the deposit holding the desired version of the key.
   * @return the key material
   * @throws IOException if key retrieval failure or timeout occurs.
   */
  @Override
  public KeyVersion getKeyVersion(String versionName) throws IOException {
    if (LOG.isDebugEnabled()) LOG.debug("Entering getKeyVersion method.");
    Preconditions.checkNotNull(versionName, "Null versionName passed to getKeyVersion.");

    byte[] material = null;

    Request request = createDepositRequest(versionName);
    String name = null;
    boolean readLocked = false;

    try {
      if (this.strongLocking) { readLock.lock(); readLocked = true; }
      DepositRelease depositRelease = keyTrusteeConnectionProxy.get(request);
      if (this.strongLocking) { readLock.unlock(); readLocked = false; }
      AnnotatedDepositRelease annotatedRelease = convert(depositRelease);
      material = IOUtils.toByteArray(depositRelease.getContents());
      name = annotatedRelease.getHandle();
    } catch (KeyTrusteeException e) {
      sendDebugInformationToLog("getKeyVersion", e);
      checkForKeyNotFound(versionName, e);
      throw new IOException("Key retrieval failed.", e);
    } catch (TimeoutException e) {
      throw new IOException("Key retrieval timed out.", e);
    } finally {
      if (readLocked) readLock.unlock();
      if (LOG.isDebugEnabled()) LOG.debug("Exiting getKeyVersion method.");
    }
    return new KeyTrusteeKeyVersion(name, versionName, material);

  }

  /**
   * Get the key material for all versions of a specific key name.
   *
   * @param name KeyTrustee handle name under which key versions are stored.
   * @return the reverse-chronologically ordered list of key material
   * @throws IOException if retrieval failure or timeout occurs.
   */
  @Override
  public List<KeyVersion> getKeyVersions(String name) throws IOException {
    if (LOG.isDebugEnabled()) LOG.debug("Entering getKeyVersions method.");
    Preconditions.checkNotNull(name, "Null name passed to getKeyVersions.");
    if (LOG.isDebugEnabled()) LOG.debug("Exiting getKeyVersions method.");
    return getKeyVersions(name, true);
  }

  private List<KeyVersion> getKeyVersions(String name, boolean enabledOnly) throws IOException {
    ArrayList<KeyVersion> keyVersionList = new ArrayList<KeyVersion>(0);
    boolean readLocked = false;

    try {
      Collection<DepositInfo> deposits = getHandleDeposits(name, enabledOnly);
      List<DepositInfo> depositList = new ArrayList(deposits);
      Set<KeyVersion> keyVersions = new LinkedHashSet<KeyVersion>(depositList.size());
      Iterator<DepositInfo> iterator = depositList.iterator();
      if (enabledOnly) {
        while (iterator.hasNext()) {
          AnnotatedDepositRelease annotatedDeposit = (AnnotatedDepositRelease)iterator.next();
          String uuid = annotatedDeposit.getUuid();
          byte[] material = IOUtils.toByteArray(annotatedDeposit.getContents());
          KeyVersion keyVersion = new KeyTrusteeKeyVersion(name, uuid, material);
          keyVersions.add(keyVersion);
        }
      } else {
        while (iterator.hasNext()) {
          DepositInfo deposit = iterator.next();
          String uuid = deposit.getUuid();
          byte[] material = null;
          if (deposit.getState().contains(DEPOSIT_STATE_ENABLED)) {
            Request request = createDepositRequest(uuid);
            if (this.strongLocking) { readLock.lock(); readLocked = true; }
            DepositRelease depositRelease = keyTrusteeConnectionProxy.get(request);
            if (this.strongLocking) { readLock.unlock(); readLocked = false; }
            material = IOUtils.toByteArray(depositRelease.getContents());
          }
          KeyVersion keyVersion = new KeyTrusteeKeyVersion(name, uuid, material);
          keyVersions.add(keyVersion);
        }
      }
      keyVersionList = new ArrayList<KeyVersion>(keyVersions);
    } catch (KeyTrusteeException e) {
      sendDebugInformationToLog("getKeyVersions", e);
      checkForKeyNotFound(name, e);
      throw new IOException("Retrieval of keys named " + name + " failed.", e);
    } catch (TimeoutException e) {
      throw new IOException("Retrieval of keys named " + name + " timed out.", e);
    } finally {
      if (readLocked) readLock.unlock();
    }
    return keyVersionList;
  }

  /**
   * Roll a new version of the given key.
   *
   * @param name     the basename of the key (KeyTrustee handle name)
   * @param material the new key material
   * @return the version name (KeyTrustee UUID) of the new version of the key
   * @throws IOException if key is not found or key could not be rolled.
   */
  @Override
  public KeyVersion rollNewVersion(String name, byte[] material) throws IOException {
    if (LOG.isDebugEnabled()) LOG.debug("Entering rollNewVersion(String,byte[]) method.");
    Preconditions.checkNotNull(name, "Null name passed to rollNewVersion.");
    Preconditions.checkNotNull(material, "Null material passed to rollNewVersion.");
    try {
      writeLock.lock();
      List<KeyVersion> keyVersions = getKeyVersions(name);
      if (keyVersions.size() < 1) throw new KeyNotFoundException("Key with name " + name + " not found in " + this);
      Metadata metadata = getMetadata(name);
      Deposit deposit = createHandleDeposit(name, material, metadata);
      String versionName = null;
      try {
        versionName = putDeposit(deposit);
      } catch (KeyTrusteeException e) {
        sendDebugInformationToLog("rollNewVersion(String,byte[])", e);
        throw new IOException("New version of key named " + name + " could not be rolled.", e);
      }
      return new KeyTrusteeKeyVersion(name, versionName, material);
    } finally {
      writeLock.unlock();
      if (LOG.isDebugEnabled()) LOG.debug("Exiting rollNewVersion(String,byte[]) method.");
    }
  }

  /**
   * Roll a new version of the given key generating the material for it.
   * <p/>
   * This implementation generates the key material and calls the
   * {@link #rollNewVersion(String, byte[])} method. This implementation should be removed
   * from the TrusteeKeyProvider once metadata storage and retrieval is implemented as the base class
   * implementation will work once this class supports metadata.
   *
   * @param name the basename of the key
   * @return the name of the new version of the key
   * @throws IOException
   */
  public KeyVersion rollNewVersion(String name) throws NoSuchAlgorithmException,
          IOException {
    // TODO This method can be removed now that metadata storage is implemented
    if (LOG.isDebugEnabled()) LOG.debug("Entering rollNewVersion(String) method.");
    // Metadata meta = new KeyTrusteeMetadata(DEFAULT_CIPHER, DEFAULT_BITLENGTH, "", new Date(), 1);
    /*
    byte[] material = generateKey(DEFAULT_BITLENGTH, DEFAULT_CIPHER);
    return rollNewVersion(name, material);
    */
    Preconditions.checkNotNull(name, "Null name passed to rollNewVersion.");
    boolean readLocked = false;
    try {
      readLock.lock();
      readLocked = true;
      List<KeyVersion> keyVersions = getKeyVersions(name);
      if (keyVersions.size() < 1) throw new KeyNotFoundException("Key with name " + name + " not found in " + this);
      readLock.unlock();
      readLocked = false;
      return super.rollNewVersion(name);
    } finally {
      if (readLocked) readLock.unlock();
      if (LOG.isDebugEnabled()) LOG.debug("Exiting rollNewVersion(String) method.");
    }
  }

  /**
   * Get the key names for all keys.
   *
   * @return the list of key names (KeyTrustee handle names)
   * @throws IOException if retrieval fails.
   */
  @Override
  public List<String> getKeys() throws IOException {
    if (LOG.isDebugEnabled()) LOG.debug("Entering getKeys method.");
    ArrayList<String> handleList = new ArrayList<String>();

    try {
      Set<String> handles = new HashSet<String>();
      Collection<DepositInfo> deposits = getAllDeposits();
      Iterator<DepositInfo> iterator = deposits.iterator();
      while (iterator.hasNext()) {
        DepositInfo deposit = iterator.next();
        String handle = deposit.getHandle();
        if (null != handle || !(handle.equals(""))) {
          handles.add(handle);
        }
      }
      handleList = new ArrayList<String>(handles);
    } catch (KeyTrusteeException e) {
      sendDebugInformationToLog("getKeys", e);
      throw new IOException("Retrieval of all keys failed.", e);
    }
    if (LOG.isDebugEnabled()) LOG.debug("Exiting getKeys method.");
    return handleList;
  }

  /**
   * Get metadata about the key.
   * @param name the basename of the key
   * @return the key's metadata or null if the key doesn't exist
   * @throws IOException
   */
  @Override
  public Metadata getMetadata(String name) throws IOException {
    if (LOG.isDebugEnabled()) LOG.debug("Entering getMetadata method.");
    Preconditions.checkNotNull(name, "Null name passed to getMetadata.");

    Metadata meta = null;
    try {
      Collection<DepositInfo> deposits = getHandleMetadata(name, true);
      if (deposits.size() > 0) {
        Iterator<DepositInfo> iterator = deposits.iterator();
        DepositInfo deposit = iterator.next();
        java.util.Map<String, String> metaMap = deposit.getMeta();
        String cipher = metaMap.remove(CIPHER_ATTRIBUTE);
        int bitLength = Integer.parseInt(metaMap.remove(BIT_LENGTH_ATTRIBUTE));
        String description = metaMap.remove(DESCRIPTION_ATTRIBUTE);
        Date created = new java.util.Date();
        try {
          created = new java.text.SimpleDateFormat("EEE MMM d HH:mm:ss z yyyy", Locale.ENGLISH)
                                                          .parse(metaMap.remove(CREATED_ATTRIBUTE));
        } catch (ParseException e) {
          throw new IOException("Could not parse stored date for " + name, e);
        }
        int versions = Integer.parseInt(metaMap.remove(VERSIONS_ATTRIBUTE));
        meta = new KeyTrusteeMetadata(cipher,bitLength,description,metaMap,created,versions);
      }
    } catch (KeyTrusteeException e) {
      if (LOG.isDebugEnabled()) {
        LOG.debug("getMetadata threw KeyTrusteeException. Possibly expected. " +
                "If unexpected an IOException will be thrown.");
      }
      try {
        checkForKeyNotFound(name, e);
      } catch (KeyNotFoundException e2) {
        return meta; // return null when key not found
      }
      throw new IOException("Could not retrieve metadata for " + name, e);
    }
    if (LOG.isDebugEnabled()) LOG.debug("Exiting getMetadata method.");
    return meta;
  }

  /**
   * Delete all versions of the given key. Note that this is a soft (reversible) delete.
   * {@link #purgeKey(String)} method provides hard (irreversible) delete.
   *
   * @param name the name (KeyTrustee handle) of the key to delete
   * @throws IOException if key could not be deleted
   */
  @Override
  public void deleteKey(String name) throws IOException {
    if (LOG.isDebugEnabled()) LOG.debug("Entering deleteKey method. Will exit from toggleKeyEnabledStatus.");
    toggleKeyEnabledStatus(Preconditions.checkNotNull(name, "Null name passed to deleteKey."), false);
  }

  /**
   * Un-delete all versions of the given key.
   * See {@link #deleteKey(String)} method for more information.
   *
   * @param name the name (KeyTrustee handle) of the key to delete
   * @throws IOException if key could not be recovered.
   */
  public void unDeleteKey(String name) throws IOException {
    if (LOG.isDebugEnabled()) LOG.debug("Entering unDeleteKey method. Will exit from toggleKeyEnabledStatus.");
    toggleKeyEnabledStatus(Preconditions.checkNotNull(name, "Null name passed to unDeleteKey."), true);
  }

  /**
   * Purge all versions of the given key. Note that this is a hard (irreversible) delete.
   * {@link #deleteKey(String)} method provides hard (reversible) delete.
   *
   * @param name the name (KeyTrustee handle) of the key to delete
   * @throws IOException if key could not be purged
   */
  public void purgeKey(String name) throws IOException {
    if (LOG.isDebugEnabled()) LOG.debug("Entering purgeKey method.");
    Preconditions.checkNotNull(name, "Null name passed to purge key.");
    List<KeyVersion> keyVersions = getKeyVersions(name, false);
    Iterator<KeyVersion> keyVersionIterator = keyVersions.iterator();
    while (keyVersionIterator.hasNext()) {
      KeyVersion keyVersion = keyVersionIterator.next();
      String uuid = keyVersion.getVersionName();
      boolean writeLocked = false;
      try {
        if (this.strongLocking) { writeLock.lock(); writeLocked = true; }
        keyTrusteeConnectionProxy.purge(uuid);
        if (this.strongLocking) { writeLock.unlock(); writeLocked = false; }
      } catch (KeyTrusteeException e) {
        sendDebugInformationToLog("purgeKey", e);
        throw new IOException("Purge of key(s) named " + name + " failed.", e);
      } finally {
        if (writeLocked) writeLock.unlock();
        if (LOG.isDebugEnabled()) LOG.debug("Exiting purgeKey method. Will exit from toggleKeyEnabledStatus.");
      }
    }
  }

  /**
   * Migrate keys from the specified key provider to this key provider.
   */
  public void migrateKeys(KeyProvider provider) throws IOException {
    if (LOG.isDebugEnabled()) LOG.debug("Entering migrateKeys method.");
    List<String> keyNames = provider.getKeys();
    List<String> summaryMessages = new ArrayList<String>();
    for (String keyName : keyNames) {
      List<KeyVersion> keyVersions = provider.getKeyVersions(keyName);
      Metadata metadata = provider.getMetadata(keyName);
      try {
        writeLock.lock();
        List<KeyVersion> existingKeyVersions = getKeyVersions(keyName);
        Metadata existingMetadata = getMetadata(keyName);
        List<KeyVersion> vettedKeyVersions = checkKeyVersions(keyVersions, existingKeyVersions, metadata, 
                                                                          existingMetadata, summaryMessages);
        for (KeyVersion keyVersion : vettedKeyVersions) {
          String name = keyVersion.getName();
          String versionName = keyVersion.getVersionName();
          byte[] material = keyVersion.getMaterial();
          migrateKey(name, material, metadata, versionName);
        }
      } finally {
        writeLock.unlock();
      }
    }
    if (summaryMessages.size() > 0) {
      if (LOG.isErrorEnabled()) {
        LOG.error("Summary of migration failures ( " + summaryMessages.size() + " ).");
        LOG.error("----------------------------------------");
        for (String message : summaryMessages) {
          LOG.error(message);
          LOG.error(System.getProperty("line.separator"));
        }
      }
    }
    if (LOG.isDebugEnabled()) LOG.debug("Exiting migrateKeys method.");
  }

  /**
   *
   * @return
   */
  public URI[] getSpecifiedProviderURIs() {
    return uris;
  }

  /**
   * Indicates whether a key with the specified name exists in this provider.
   * @param name KeyTrustee handle name under which key versions are stored.
   * @return true if key with specified name exists; false otherwise
   * @throws IOException if key retrieval fails or timeout occurs
   */
  public KeyExistsResult keyExists(String name) throws IOException {
    if (LOG.isDebugEnabled()) LOG.debug("Entering keyExists method.");
    Preconditions.checkNotNull(name, "Null name passed to keyExists.");
    List<KeyVersion> keyVersions = getKeyVersions(name, false);
    String reason = "Key does not exist.";
    if (keyVersions.size() > 0) {
      reason = "Key exists and is active.";
        KeyVersion keyVersion = keyVersions.get(0);
        byte[] material = keyVersion.getMaterial();
        if (null == material) reason = "Key exists but has been disabled. Use undelete to enable.";
        if (LOG.isDebugEnabled()) LOG.debug("Exiting keyExists method.");
        return new KeyExistsResult(true, reason);
    }
    if (LOG.isDebugEnabled()) LOG.debug("Exiting keyExists method.");
    return new KeyExistsResult(false, reason);
  }

  /**
   * Ensures that any changes to the keys are written to persistent store.
   * This method has an empty implementation and performs no operation on this provider class.
   * @throws IOException
   */
  @Override
  public void flush() throws IOException {

  }

  private KeyVersion migrateKey(String name, byte[] material, Metadata metadata, String oldVersionName) 
          throws IOException {
    Preconditions.checkNotNull(name, "Null name passed to migrateKey.");
    Preconditions.checkNotNull(material, "Null material passed to migrateKey.");
    Preconditions.checkNotNull(metadata, "Null metadata passed to migrateKey.");

    try {
      // writeLock.lock(); // locked in migrateKeys method
      List<KeyVersion> existingKeyVersions = getKeyVersions(name);
      if (LOG.isInfoEnabled()) {
        LOG.info("Number of existing key versions named " + name + ": " + existingKeyVersions.size());
      }
      Deposit deposit = migrateHandleDeposit(name, material, metadata, oldVersionName, existingKeyVersions.size());
      String versionName = null;
      try {
        versionName = putDeposit(deposit);
      } catch (KeyTrusteeException e) {
        sendDebugInformationToLog("migrateKey", e);
        throw new IOException("Migration of key named " + name + " failed.", e);
      }
      if (LOG.isInfoEnabled()) {
        LOG.info("Migrated key named " + name + ". New version name: " + versionName + 
                                              ". Old version name: " + oldVersionName + ".");
      }
      return new KeyTrusteeKeyVersion(name, versionName, material);
    } finally {
      //writeLock.unlock(); // unlocked in migrateKeys method
    }
  }

  private Deposit migrateHandleDeposit(String name, byte[] material, Metadata metadata, 
                                              String originalVersionName, int versionCount) {
    int metadataSize = 5;
    Map<String,String> attributes = metadata.getAttributes();
    if (null != attributes) metadataSize += attributes.size();
    Map<String, String> map = new HashMap<String,String>(metadataSize);
    map.put(VERSIONS_ATTRIBUTE, String.valueOf(versionCount + 1));
    map.put(CREATED_ATTRIBUTE, String.valueOf(metadata.getCreated()));
    map.put(BIT_LENGTH_ATTRIBUTE, String.valueOf(metadata.getBitLength()));
    map.put(CIPHER_ATTRIBUTE, metadata.getCipher());
    if (null != metadata.getDescription()) map.put(DESCRIPTION_ATTRIBUTE, metadata.getDescription());
    if (null != attributes) map.putAll(attributes);
    map.put("original_id", originalVersionName);
    return createHandleDeposit(name, material, map);
  }

  /*
   * Helper method for constructor only.
   */
  private TrusteeClientConnectionPool initializeConnectionPool(File keyTrusteeConfDir, 
                                                               TrusteeKeyProviderConfiguration tkpConfig)
                                                                    throws KeyTrusteeException, IOException {
    PooledClientConnectionFactory pooledClientConnectionFactory = 
              new PooledClientConnectionFactory(keyTrusteeConfDir, tkpConfig);
    return pooledClientConnectionFactory.getKeyTrusteeConnectionPool();
  }
  
  private boolean isRegistered(ClientFactory keyTrusteeClientFactory, File keyTrusteeConfDir) throws IOException {
    try {
      ClientInfo keyTrusteeClient = keyTrusteeClientFactory.openClient(keyTrusteeConfDir);
      ServerInfo keyTrusteeServer = keyTrusteeClient.getServerInfo();
      if (null == keyTrusteeServer) return false;
      ClientConnection zTrusteeConnection = keyTrusteeClientFactory.connect(keyTrusteeClient, keyTrusteeServer);
      return zTrusteeConnection.isRegistered();
    } catch (KeyTrusteeException e) {
      sendDebugInformationToLog("isRegistered", e);
      throw new IOException("KeyProvider failed during registration check.", e);
    }
  }

  private void checkForKeyNotFound(String name, KeyTrusteeException e) throws KeyNotFoundException {
    if (null != e.getCause() && e.getCause() instanceof HTTPException) {
      HTTPException httpException = (HTTPException)e.getCause();
      if (httpException.getStatusCode() == 404) {
        throw new KeyNotFoundException("Key " + name + " not found.", e);
      }
    } else {
      if (null != e.getCause() && e.getCause() instanceof KeyTrusteeException) {
        checkForKeyNotFound(name, (KeyTrusteeException)e.getCause());
      }
    }
  }

  private static void removeConfDir(File keyTrusteeConfDir) {
    keyTrusteeConfDir.delete();
  }

  private void assertKeyAccess(String keyName, String keyGroup) throws IOException {
    String userName = UserGroupInformation.getCurrentUser().getShortUserName();
    if (!groupsMapping.getGroups(userName).contains(keyGroup)) {
      throw new AuthorizationException(MessageFormat.format(
              "User:{0} not allowed to access key ''{1}'', not in ''{2}'' group",
              userName, keyName, keyGroup));
    }
  }

  private static void assertKeyTrusteeUUID(String uuid) throws IOException {
    if (!(verifyKeyTrusteeUUID(uuid) || verifyJKSUUID(uuid))) throw new IOException("Invalid uuid: " + uuid);
  }

  private static boolean verifyKeyTrusteeUUID(String uuid) {
    // zTrustee UUID is a 43 character base62 random UUID
    // For now, just verifying length
    return uuid.length() == 43;
  }

  private static boolean verifyJKSUUID(String id) {
    // JKS ID is an string ending in "@N" where N is an integer.
    return id.matches("^.+?@\\d$");
  }



  private List<KeyVersion> checkKeyVersions(List<KeyVersion> proposedKeyVersions, List<KeyVersion> existingKeyVersions,
                                            Metadata proposedMetadata, Metadata existingMetadata, 
                                            List<String> summaries) {
    ArrayList<KeyVersion> vettedKeyVersions = new ArrayList<KeyVersion>();
    for (KeyVersion proposedKeyVersion : proposedKeyVersions) {
      vettedKeyVersions.add(proposedKeyVersion);
      for (KeyVersion existingKeyVersion : existingKeyVersions) {
        String proposedVersionName = proposedKeyVersion.getVersionName();
        String existingVersionName = existingKeyVersion.getVersionName();
        Map<String, String> existingAttributes = existingMetadata.getAttributes();
        String originalId = existingAttributes.get("original_id");
        if (proposedVersionName.equalsIgnoreCase(existingKeyVersion.getVersionName()) ||
                proposedVersionName.equalsIgnoreCase(originalId)) {
          vettedKeyVersions.remove(proposedKeyVersion);
          String message = "Found proposed key version name (" + proposedKeyVersion.getVersionName() + 
                                                                      ") in target provider already. Skipping.";
          String detail = "Existing key version name: " + existingKeyVersion.getVersionName() + 
                                                                  ". Existing original_id: " + originalId + ".";
          if (LOG.isWarnEnabled()) {
            LOG.warn(message);
            LOG.warn(detail);
          }
          byte[] proposedMaterial = proposedKeyVersion.getMaterial();
          byte[] existingMaterial = existingKeyVersion.getMaterial();
          String inconsistency = "Existing material in provider does not match proposed key material.";
          String inconsistencyDetail = "Existing key material: " + existingMaterial + 
                                                  ". Proposed key material: " + proposedMaterial + ".";
          if (!(Arrays.equals(proposedMaterial, existingMaterial))) {
            if (LOG.isWarnEnabled()) {
              LOG.warn(inconsistency);
              LOG.warn(inconsistencyDetail);
            }
            String sep = System.getProperty("line.separator");
            summaries.add( message + sep + detail + sep + message + sep + detail );
          }
        }
      }
    }
    return vettedKeyVersions;
  }

  /**
   * Returns all enabled deposits for the client identity this provider is using to connect to KeyTrustee server.
   */
  private Collection<DepositInfo> getAllDeposits() throws KeyTrusteeException {
    return getAllDeposits(true);
  }

  /**
   * Returns all deposits for the client identity this provider is using to connect to KeyTrustee server.
   * @param enabledOnly if true, return only Deposits whose state is Enabled.
   */
  private Collection<DepositInfo> getAllDeposits(boolean enabledOnly) throws KeyTrusteeException {
    return getAllDeposits(QUERY_ALL_ENABLED_DEPOSITS);
  }

  private Collection<DepositInfo> getHandleDeposits(String handleName, boolean enabledOnly)
          throws IOException, KeyTrusteeException {
    if (enabledOnly) {
      Request request = createHandleDepositRequest(handleName);
      boolean readLocked = false;

      try {
        if (this.strongLocking) { readLock.lock(); readLocked = true; }
        DepositRelease[] depositReleases = keyTrusteeConnectionProxy.getHandleDeposits(request);
        if (this.strongLocking) { readLock.unlock(); readLocked = false; }
        AnnotatedDepositRelease[] annotatedReleases = convert(depositReleases);
        return new ArrayList<DepositInfo>(Arrays.asList(annotatedReleases));
      } catch (TimeoutException e) {
        throw new IOException("Key retrieval timed out.", e);
      } finally {
        if (readLocked) readLock.unlock();
      }
    } else {
      DepositSearch query = new DepositSearch.Builder().handle(true).keyword(handleName).enabled(enabledOnly).build();
      return getAllDeposits(query);
    }
  }

  private AnnotatedDepositRelease[] convert(DepositRelease[] releases) throws IOException {
    AnnotatedDepositRelease[] annotatedReleases = new AnnotatedDepositRelease[releases.length];
    for (int i=0; i < releases.length; i++) {
      annotatedReleases[i] = convert(releases[i]);
    }
    return annotatedReleases;
  }

  private AnnotatedDepositRelease convert(DepositRelease release) throws IOException {
    if (release instanceof AnnotatedDepositRelease) {
      AnnotatedDepositRelease annotatedRelease = (AnnotatedDepositRelease) release;
      return annotatedRelease;
    } else {
      throw new IOException("Expected AnnotatedDepositRelease but found DepositRelease.");
    }
  }

  private Collection<DepositInfo> getHandleMetadata(String handleName, boolean enabledOnly)
          throws IOException, KeyTrusteeException {
    if (enabledOnly) {
      Request request = createHandleDepositRequest(handleName);
      boolean readLocked = false;
      try {
        if (this.strongLocking) { readLock.lock(); readLocked = true; }
        DepositRelease[] depositReleases = keyTrusteeConnectionProxy.getHandleDeposits(request);
        if (this.strongLocking) { readLock.unlock(); readLocked = false; }
        AnnotatedDepositRelease[] annotatedReleases = convert(depositReleases);
        return new ArrayList<DepositInfo>(Arrays.asList(annotatedReleases));
      } catch (TimeoutException e) {
        throw new IOException("Key retrieval timed out.", e);
      } finally {
        if (readLocked) readLock.unlock();
      }
    } else {
      DepositSearch query =
              new DepositSearch.Builder().handle(true).keyword(handleName).enabled(enabledOnly).meta(true).build();
      return getAllDeposits(query);
    }
  }

  /**
   * Returns deposits for the client identity this provider is using to connect to KeyTrustee server.
   * @param query query to use in searching deposits
   */
  private Collection<DepositInfo> getAllDeposits(DepositSearch query) throws KeyTrusteeException {
    boolean readLocked = false;
    try {
      if (this.strongLocking) {
        readLock.lock();
        readLocked = true;
      }
      Collection<DepositInfo> deposits = keyTrusteeConnectionProxy.getDeposits(query);
      if (this.strongLocking) {
        readLock.unlock();
        readLocked = false;
      }
      return deposits;
    } finally {
      if (readLocked) readLock.unlock();
    }
  }

  private String getDepositInfoByUUID(String versionName) throws KeyTrusteeException, IOException {
    String name;DepositSearch.Builder builder = new DepositSearch.Builder();
    builder.keyword(versionName);
    DepositSearch query = builder.build();
    Collection<DepositInfo> depositInfos = getAllDeposits(QUERY_ALL_ENABLED_DEPOSITS);
    if (depositInfos.size() > 1) throw new IOException("Found multiple deposits for uuid " + versionName);
    if (depositInfos.size() < 1) throw new IOException("Found no deposit for uuid " + versionName);
    Iterator<DepositInfo> iterator = depositInfos.iterator();
    DepositInfo depositInfo = iterator.next();
    name = depositInfo.getHandle();
    return name;
  }

  private Request createDepositRequest(String uuid) throws IOException {
    Request.Builder requestBuilder = new Request.Builder();
    assertKeyTrusteeUUID(uuid);
    requestBuilder.uuid(uuid);
    requestBuilder.include_meta(true);
    return requestBuilder.build();
  }

  private Request createHandleDepositRequest(String handleName) throws IOException {
    Request.Builder requestBuilder = new Request.Builder();
    requestBuilder.handle(handleName);
    requestBuilder.include_meta(true);
    return requestBuilder.build();
  }

  private Request createOriginalIdDepositRequest(String originalId) throws IOException {
    Request.Builder requestBuilder = new Request.Builder();
    requestBuilder.originalId(originalId);
    requestBuilder.include_meta(true);
    return requestBuilder.build();
  }

  /**
   * Creates a keyTrustee deposit object suitable for use in a ClientConnection.put operation
   * from Hadoop Key Provider key name and material.
   *
   * @param name     Handle name for the deposit.
   * @param material Content for the deposit.
   * @param options Options for deposit (including arbitrary key/value attributes)
   * @return com.cloudera.keytrustee.Deposit
   */
  private Deposit createHandleDeposit(String name, byte[] material, Options options) {
    Map<String, String> metadata = new HashMap<String,String>();
    if (null != options) {
      int metadataSize = 5; // prepare to store static attributes only
      int bitLength = options.getBitLength();
      String cipher = options.getCipher();
      String description = options.getDescription();

      Map<String, String> attributes = options.getAttributes();
      if (null != attributes) metadataSize += attributes.size(); // prepare to store static and arbitrary attributes
      metadata = new HashMap<String, String>(metadataSize);
      metadata.put(VERSIONS_ATTRIBUTE, "1");
      metadata.put(CREATED_ATTRIBUTE, new java.text.SimpleDateFormat("EEE MMM d HH:mm:ss z yyyy", Locale.ENGLISH)
                                                                                      .format(new java.util.Date()));
      metadata.put(BIT_LENGTH_ATTRIBUTE, String.valueOf(bitLength));
      metadata.put(CIPHER_ATTRIBUTE, cipher);
      if (null != description) metadata.put(DESCRIPTION_ATTRIBUTE, description);
      if (null != attributes) metadata.putAll(attributes);
    } else {
      if (LOG.isWarnEnabled())
        LOG.warn("Null options passed to createHandleDeposit. This should throw an exception before release.");
    }

    return createHandleDeposit(name, material, metadata);
  }

  private Deposit createHandleDeposit(String name, byte[] material, Metadata metadata) {
    int metadataSize = 5;
    Map<String,String> attributes = metadata.getAttributes();
    if (null != attributes) metadataSize += attributes.size();
    Map<String, String> map = new HashMap<String,String>(metadataSize);
    map.put(VERSIONS_ATTRIBUTE, String.valueOf(metadata.getVersions() + 1));
    map.put(CREATED_ATTRIBUTE, String.valueOf(metadata.getCreated()));
    map.put(BIT_LENGTH_ATTRIBUTE, String.valueOf(metadata.getBitLength()));
    map.put(CIPHER_ATTRIBUTE, metadata.getCipher());
    if (null != metadata.getDescription()) map.put(DESCRIPTION_ATTRIBUTE, metadata.getDescription());
    if (null != attributes) map.putAll(attributes);
    return createHandleDeposit(name, material, map);
  }

  private Deposit createHandleDeposit(String name, byte[] material, Map<String,String> map) {
    Deposit.Builder depositBuilder = new Deposit.Builder();
    depositBuilder.handle(name);
    depositBuilder.content(new ByteArrayInputStream(material));
    depositBuilder.metadata(map);
    return depositBuilder.build();
  }

  private String putDeposit(Deposit deposit) throws KeyTrusteeException {
    boolean writeLocked = false;
    try {
      if (this.strongLocking) {
        writeLock.lock();
        writeLocked = true;
      }
      DepositInfo info = keyTrusteeConnectionProxy.put(deposit);
      if (this.strongLocking) {
        writeLock.unlock();
        writeLocked = false;
      }
      String uuid = info.getUuid();
      return uuid;
    } finally {
      if (writeLocked) writeLock.unlock();;
    }
  }

  private List<DepositInfo> getSortedDepositInfoList(Collection<DepositInfo> deposits) {
    List<DepositInfo> depositList = new ArrayList<DepositInfo>(deposits.size());
    depositList.addAll(deposits);
    Comparator<DepositInfo> depositInfoComparator = new Comparator<DepositInfo>() {

      public int compare(DepositInfo deposit1, DepositInfo deposit2) {
        return deposit2.getCreation().compareTo(deposit1.getCreation());
      }
    };
    Collections.sort(depositList, depositInfoComparator);
    return depositList;
  }

  private void toggleKeyEnabledStatus(String name, boolean enable) throws IOException {
    try {
      writeLock.lock();
      boolean enabledonly = false;
      List<KeyVersion> keyVersions = getKeyVersions(name, enabledonly);
      String message = "Key with name " + name + " not found in " + this + ". ";
      if (keyVersions.size() < 1) throw new IOException(message);
      Iterator<KeyVersion> keyVersionIterator = keyVersions.iterator();
      while (keyVersionIterator.hasNext()) {
        KeyVersion keyVersion = keyVersionIterator.next();
        String uuid = keyVersion.getVersionName();
        byte[] material = keyVersion.getMaterial();
        try {
          if (enable) {
            keyTrusteeConnectionProxy.enable(uuid);
          } else {
            if (null == material) throw new IOException("Key named " + name + 
                    " was already deleted but is disabled. Use purge to destroy all traces or undelete to reactivate.");
            keyTrusteeConnectionProxy.disable(uuid);
          }
        } catch (KeyTrusteeException e) {
          sendDebugInformationToLog("toggleKeyEnabledStatus", e);
          if (enable) {
            throw new IOException("Un-deletion of key(s) named " + name + " failed.", e);
          } else {
            throw new IOException("Deletion of key(s) named " + name + " failed.", e);
          }
        }
      }
    } finally {
      writeLock.unlock();
      String methodName = "deleteKey method.";
      if (enable) methodName = "unDeleteKey method.";
      if (LOG.isDebugEnabled()) LOG.debug("Exiting toggleKeyEnabledStatus method. Entered from " + methodName);
    }
  }
  
  private void sendDebugInformationToLog(String methodName, KeyTrusteeException e) {
    if (LOG.isDebugEnabled()) LOG.debug(methodName + " threw KeyTrusteeException: " + e.getMessage());
  }

    public int getIdleConnectionCount() {
      ClientConnectionInvocationHandler handler = 
              (ClientConnectionInvocationHandler)java.lang.reflect.Proxy.getInvocationHandler(keyTrusteeConnectionProxy);
      return handler.getIdleConnectionCount();
    }

    public int getActiveConnectionCount() {
      ClientConnectionInvocationHandler handler = 
              (ClientConnectionInvocationHandler)java.lang.reflect.Proxy.getInvocationHandler(keyTrusteeConnectionProxy);
      return handler.getActiveConnectionCount();
    }

  /**
   * The factory to create TrusteeKeyProvider, which is used by the
   * ServiceLoader.
   */
  public static class Factory extends KeyProviderFactory {

    @Override
    public KeyProvider createProvider(URI providerName, Configuration conf) throws IOException {
      Preconditions.checkNotNull(providerName, "Null providerName passed to KeyProvider.Factory.createProvider.");
      if (SCHEME_NAME.equals(providerName.getScheme())) {
        return new TrusteeKeyProvider(providerName, conf);
      }
      return null;
    }
  }

  public static class KeyTrusteeKeyVersion extends KeyVersion {
    public KeyTrusteeKeyVersion(String name, String versionName, byte[] material) {
      super(name, versionName, material);
    }
  }

  public static class KeyTrusteeMetadata extends Metadata {
    public KeyTrusteeMetadata(String cipher, int bitLength, String description,
                              Map<String, String> attributes, Date created, int versions) {
      super(cipher, bitLength, description, attributes, created, versions);
    }
  }

  public class KeyExistsResult {
    private boolean existing;
    private String reason;

    public KeyExistsResult(boolean exists, String reason) {
      this.existing = exists;
      this.reason = reason;
    }

    public boolean isExisting() {
      return existing;
    }

    public String getReason() {
      return reason;
    }
  }

  public class DuplicateKeyException extends IOException {

    public DuplicateKeyException(String message) {
      super(message);
    }

    public DuplicateKeyException(String message, Throwable cause) {
      super(message, cause);
    }
  }

  public class KeyNotFoundException extends IOException {

    public KeyNotFoundException(String message) {
      super(message);
    }

    public KeyNotFoundException(String message, Throwable cause) {
      super(message, cause);
    }
  }
}
