/*
 * Copyright (c) 2014 Cloudera, Inc. All rights reserved.
 */

package com.cloudera.keytrustee.test;

import com.cloudera.keytrustee.TrusteeKeyProvider;
import com.cloudera.keytrustee.TrusteeKeyProviderConfiguration;
import com.cloudera.keytrustee.util.Environment;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.crypto.key.KeyProvider;
import org.junit.Rule;
import org.junit.rules.TemporaryFolder;

import java.io.IOException;
import java.security.NoSuchAlgorithmException;

/**
 * Shared logic for testing is factored out into this class so it can be used by multiple Test classes.
 */
public class TestHelper {
  //see http://docs.oracle.com/javase/6/docs/technotes/guides/security/StandardNames.html#AlgorithmParameters
  protected static final String DEFAULT_ALGORITHM = "AES";
  protected static final int DEFAULT_BITLENGTH = KeyProvider.DEFAULT_BITLENGTH; // 256
  private static final String[] HADOOP_CONF =  { "kms-site.xml", "kts-site.xml" };
  @Rule
  public TemporaryFolder folder = new TemporaryFolder();

  protected TrusteeKeyProvider registerNewProvider() {
    return registerNewProvider(HADOOP_CONF);
  }

  protected TrusteeKeyProvider registerNewProvider(String[] hadoopConfigurations) {
    String tempPath = folder.newFolder(".keytrustee-" + getUniqueSuffix()).getAbsolutePath();
    //String tempPath = "/tmp/.keytrustee-"+ getUniqueSuffix();
    return registerOrLoadProvider(hadoopConfigurations, tempPath);
  }

  protected TrusteeKeyProvider registerOrLoadProvider(String[] hadoopConfigurations, String tempPath) {
    MockEnvironment mockEnv =
            new MockEnvironment(tempPath);
    Configuration configuration = new Configuration(false); // don't load any default configuration files
    for (String resource : hadoopConfigurations) {
      configuration.addResource(resource);
    }
    try {
      TrusteeKeyProvider tkp = new TrusteeKeyProvider(mockEnv, null, configuration);
      return tkp;
    } catch (IOException e) {
      throw new RuntimeException(e);
    }
  }

  protected Configuration getDefaultConfiguration() {
    Configuration conf = new Configuration(false);
    conf.set(KeyProvider.DEFAULT_CIPHER_NAME, KeyProvider.DEFAULT_CIPHER);
    conf.set(KeyProvider.DEFAULT_BITLENGTH_NAME, String.valueOf(KeyProvider.DEFAULT_BITLENGTH));
    return conf;
  }

  protected byte[] getKeyMaterial(String algorithm, int size) throws NoSuchAlgorithmException {
    javax.crypto.KeyGenerator keygenerator = javax.crypto.KeyGenerator.getInstance(algorithm);
    keygenerator.init(size);
    return keygenerator.generateKey().getEncoded();
  }

  protected String getUniqueSuffix() {
    return String.valueOf(System.currentTimeMillis());
  }

  private static class MockEnvironment extends Environment {
    private static java.util.HashMap<String, String> map = new java.util.HashMap<String, String>();
    private final Environment environment = new Environment();

    MockEnvironment(String zTrusteeConfDirPath) {
      map.put(TrusteeKeyProviderConfiguration.KEYTRUSTEE_CONF_DIR_ENV_VAR, zTrusteeConfDirPath);
    }

    public boolean containsKey(String key) {
      if (map.containsKey(key)) return true;
      return environment.containsKey(key);
    }

    public String get(String key) {
      if (map.containsKey(key)) return map.get(key);
      return environment.get(key);
    }

    public boolean getBoolean(String key) {
      if (map.containsKey(key)) return Boolean.parseBoolean(map.get(key));
      return environment.getBoolean(key);
    }
  }
}
